from .generic_dataset import GenericDataset
from .grib_dataset import GribDataset
from .netcdf_dataset import NetcdfDataset
from .predictor_dataset import PredictorDataset

__all__ = ('GenericDataset', 'GribDataset', 'NetcdfDataset', 'PredictorDataset')



import os

import matplotlib.pyplot as plt
import pandas as pd

from atmoswing_toolbox.parsers.optimized_parameters import ParametersArray


class MonteCarloPlot:
    """Plotting the results of the Monte Carlo analysis"""

    def __init__(self, file, output_path=''):
        self.fig = None
        self.results = ParametersArray(file)
        self.results.load()
        self.other_results = []
        self.other_results_score = []
        self.other_results_markers = []
        self.other_results_colors = []
        self.output_path = output_path
        self._do_print = False
        self._step = 0
        self._ptor = 0

    def show(self):
        plt.ion()
        self._loop_structure()

    def print(self):
        if not self.output_path:
            raise Exception('Output path not provided')
        self._do_print = True
        plt.ioff()
        self._loop_structure()

    def add_param(self, file, marker='+', period='valid', color=''):
        other_result = ParametersArray(file)
        other_result.load()

        # Make some checks
        if len(other_result.get_anbs(0)) != 1:
            raise Exception(f'The parameters file should contain exactly 1 set '
                            f'(the file has {other_result.get_anbs(0)}).')
        if len(self.results.struct) != len(other_result.struct):
            raise Exception(f'The number of steps is different between files '
                            f'({len(self.results.struct)} vs '
                            f'{len(other_result.struct)}).')
        for i_step, step in enumerate(self.results.struct):
            if step != other_result.struct[i_step]:
                raise Exception(f'The number of predictors is different between files '
                                f'({step} vs {other_result.struct[i_step]}).')
        # Store data
        if (period == 'valid') or (period == 'validation'):
            self.other_results_score.append(other_result.get_valid_score())
            if color == '':
                color = 'blue'
        elif (period == 'calib') or (period == 'calibration'):
            self.other_results_score.append(other_result.get_calib_score())
            if color == '':
                color = 'red'
        else:
            raise Exception(f'The provided period {period} is not recognized.')

        self.other_results.append(other_result)
        self.other_results_markers.append(marker)
        self.other_results_colors.append(color)

    def _loop_structure(self):
        for i_step, step in enumerate(self.results.struct):
            title = f'Level {i_step + 1}'
            self.make_plot('anb', xlabel='Number of analogues', title=title)
            self._print_or_show(title, 'Number of analogues')
            for self._ptor in range(0, step):
                title = f'Level {i_step + 1} - ' \
                        f'{self.results.get_variable(i_step, self._ptor)}' \
                        f'{self.results.get_level(i_step, self._ptor)} ' \
                        f'{self.results.get_time(i_step, self._ptor)}h'
                self.make_plot('xmin', xlabel='Minimum longitude [°]', title=title)
                self._print_or_show(title, 'Minimum longitude [°]')
                self.make_plot('xmax', xlabel='Maximum longitude [°]', title=title)
                self._print_or_show(title, 'Maximum longitude [°]')
                self.make_plot('ymin', xlabel='Minimum latitude [°]', title=title)
                self._print_or_show(title, 'Minimum latitude [°]')
                self.make_plot('ymax', xlabel='Maximum latitude [°]', title=title)
                self._print_or_show(title, 'Maximum latitude [°]')

    def make_plot(self, var, xlabel, title):
        self.fig = plt.figure(figsize=(5, 4))
        score = self.results.get_calib_scores()

        # Get MC values and plot
        values = None
        if var == 'anb':
            values = self.results.get_anbs(self._step)
        elif var == 'xmin':
            values = self.results.get_xmins(self._step, self._ptor)
        elif var == 'xmax':
            values = self.results.get_xmaxs(self._step, self._ptor)
        elif var == 'ymin':
            values = self.results.get_ymins(self._step, self._ptor)
        elif var == 'ymax':
            values = self.results.get_ymaxs(self._step, self._ptor)
        plt.scatter(values, score, c='', edgecolors=(0.2, 0.2, 0.2, 0.7),
                    linewidths=0.5, s=30)

        # Other values (ex. from calibration or GAs)
        for idx, res in enumerate(self.other_results):
            other_value = None
            if var == 'anb':
                other_value = res.get_anbs(self._step)
            elif var == 'xmin':
                other_value = res.get_xmins(self._step, self._ptor)
            elif var == 'xmax':
                other_value = res.get_xmaxs(self._step, self._ptor)
            elif var == 'ymin':
                other_value = res.get_ymins(self._step, self._ptor)
            elif var == 'ymax':
                other_value = res.get_ymaxs(self._step, self._ptor)
            other_score = self.other_results_score[idx]
            score = score.append(pd.Series(other_score))
            marker = self.other_results_markers[idx]
            color = self.other_results_colors[idx]
            plt.plot(other_value, other_score, marker=marker, color=color)

        # Formatting
        xmargin = 0.02 * (values.max() - values.min())
        if xmargin == 0:
            xmargin = 0.1
        plt.xlim(values.min() - xmargin, values.max() + xmargin)
        if self.results.score == 'CRPSS':
            plt.ylim(score.quantile(.75), 1.03 * score.max())
        elif self.results.score == 'CRPS':
            plt.ylim(0.97 * score.min(), score.quantile(.25))
        else:
            plt.ylim(0.97 * score.min(), score.quantile(.25))
        plt.xlabel(xlabel)
        plt.ylabel(self.results.score)
        plt.title(title)
        self.fig.tight_layout()

    def _print_or_show(self, title, xlabel):
        if self._do_print:
            filename = title + '_' + xlabel
            filename = filename.lower()
            filename = filename.replace('-', '')
            filename = filename.replace(' [°]', '')
            filename = filename.replace('  ', ' ')
            filename = filename.replace(' ', '_')
            filename = filename.replace('/', '_')
            filename = filename.replace('minimum', 'min')
            filename = filename.replace('maximum', 'max')
            filename = filename.replace('longitude', 'lon')
            filename = filename.replace('latitude', 'lat')
            filename = filename.replace('number_of', 'nb')
            self.fig.savefig(os.path.join(self.output_path, filename + '.pdf'))
            self.fig.savefig(os.path.join(self.output_path, filename + '.png'), dpi=300)
            plt.close(self.fig)
        else:
            plt.show()



import glob
import os


def cleanup_empty_log_files(base_path):
    def log_file_is_empty(file):
        if os.stat(file).st_size == 0:
            return True
        with open(file) as fi:
            if 'Optimization has already converged.' in fi.read():
                return True
        return False

    for x in os.listdir(base_path):
        path = os.path.join(base_path, x)
        logs = glob.glob(path + '/AtmoSwingOptimizer*.log')
        for f in logs:
            if log_file_is_empty(f):
                os.remove(f)
                print(f'{f} removed')


def cleanup_duplicate_ini_files(base_path):
    for x in os.listdir(base_path):
        path = os.path.join(base_path, x)
        files = glob.glob(path + '/AtmoSwing-*.ini')
        for f in files:
            os.remove(f)
            print(f'{f} removed')


def cleanup_old_generations(base_path):
    for x in os.listdir(base_path):
        path = os.path.join(base_path, x, 'results')
        files_generations = glob.glob(path + '/*generations.txt.gz')
        files_generations.sort()
        files_operators = glob.glob(path + '/*operators.txt.gz')
        files_operators.sort()
        for f in files_generations[:-1]:
            os.remove(f)
            print(f'{f} removed')
        for f in files_operators[:-1]:
            os.remove(f)
            print(f'{f} removed')



import glob
import os
from collections import Counter

import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from atmoswing_toolbox.parsers.optimized_parameters import ParametersArray


class GAsVariablesPlot:
    """Plotting the results of the Monte Carlo analysis"""

    def __init__(self, base_dir, output_path=''):
        self.fig = None
        self.legend = None
        self.base_dir = base_dir
        self.output_path = output_path
        self.marker_size_max = 50
        self.marker_size_range = 0.05
        self.marker_size_on_weight = False
        self.filter_min_weight = 0.05
        self.variables_importance_nb = 30
        self.marker_alpha = 1
        self.struct = []
        self.data = []
        self.vars = []
        self.vars_default = [
            'Z200', 'Z300', 'Z400', 'Z500', 'Z600', 'Z700', 'Z800', 'Z850', 'Z900',
            'Z925', 'Z950', 'Z1000',
            'ZA1000',
            'PV/Z',
            'U200', 'U300', 'U400', 'U500', 'U600', 'U700', 'U800', 'U1000', 'U10m',
            'PT/U285', 'PT/U300', 'PT/U315', 'PT/U330', 'PT/U350', 'PT/U370', 'PT/U395',
            'PV/U',
            'V400', 'V500', 'V600', 'V700', 'V800', 'V900', 'V950', 'V1000', 'V10m',
            'PT/V285', 'PT/V315', 'PT/V330', 'PT/V350', 'PT/V370', 'PT/V395',
            'PV/V',
            'PT/PRES285', 'PV/PRES',
            'SLP',
            'W200', 'W300', 'W400', 'W500', 'W600', 'W700', 'W800', 'W850', 'W900',
            'W950', 'W1000',
            'D300', 'D400', 'D600', 'D800', 'D850', 'D900', 'D950', 'D1000', 'D2m',
            'PT/D285', 'PT/D315', 'PT/D330', 'PT/D350', 'PT/D370',
            'VO300', 'VO500', 'VO600', 'VO700', 'VO900',
            'PV300', 'PV400', 'PV500', 'PV600', 'PV700', 'PV800', 'PV850', 'PV900',
            'PV950', 'PV1000',
            'PT/PV285', 'PT/PV330', 'PT/PV350', 'PT/PV370', 'PT/PV395',
            'PT/MONT285', 'PT/MONT300', 'PT/MONT330', 'PT/MONT350', 'PT/MONT370',
            'RH500', 'RH600', 'RH700', 'RH800', 'RH850', 'RH900', 'RH950', 'RH1000',
            'MI500', 'MI600', 'MI700', 'MI800', 'MI850', 'MI900', 'MI950', 'MI1000',
            'SH500', 'SH600', 'SH700',
            'PT/SH315', 'PT/SH330', 'PT/SH350', 'PT/SH370', 'PT/SH395',
            'TCW', 'CWAT',
            'IE',
            'T400', 'T500', 'T600', 'T700', 'T800', 'T850', 'T900', 'T950', 'T2m',
            'PV/PT',
            'DEG0L',
            'SSR', 'SSRD', 'STR', 'STRD',
            'SLHF', 'SSHF',
            'TSR', 'TTR',
            'CAPE',
            'TCC', 'CC500', 'CC600', 'CC700', 'CC800', 'CC850', 'CC1000',
            'LCC', 'SD'
        ]
        self.use_vars_default = True
        self.stations = []
        self.time_step = 6
        self.crit = ['RMSE', 'S0', 'S1', 'S2', 'MD', 'DSD', 'DMV']
        self.colors = plt.get_cmap('tab10').colors
        self.markers = ['o', 'v', 's', 'P', '^', '<', '>', '8', 'p', '*', 'h', 'H', 'D',
                        'd', 'X']
        self.files = glob.glob(base_dir + '/**/*best_individual.txt', recursive=True)

    def show_scatter(self, show_criteria=True):
        plt.ion()
        if show_criteria:
            self._set_criteria_color()
        self._set_marker_size()
        self._make_scatter_plot(show_criteria)
        plt.show()

    def print_scatter(self, filename, show_criteria=True):
        if not self.output_path:
            raise Exception('Output path not provided')
        plt.ioff()
        if show_criteria:
            self._set_criteria_color()
        self._set_marker_size()
        self._make_scatter_plot(show_criteria)
        self._print(filename)

    def print_syntheses(self, filename, show_criteria=True, target_lons=None,
                        target_lats=None, sort_by_selection=True):
        if not self.output_path:
            raise Exception('Output path not provided')
        plt.ioff()
        if show_criteria:
            self._set_criteria_color()
        self._make_syntheses_plot(show_criteria, target_lons, target_lats,
                                  sort_by_selection)
        self._print(filename)

    def load(self):
        self._parse_results()
        self._list_stations()
        self._drop_bad_scores()
        self._list_variables()
        self._add_variable_index()

    def _parse_results(self):
        data = []
        resCheck = ParametersArray(self.files[0])
        resCheck.load()
        self.struct = resCheck.struct
        labels_slct = []

        # Create labels
        for step, ptors in enumerate(self.struct):
            labels_slct.append(f'anb_{step}')
            for ptor in range(ptors):
                labels_slct.append(f'var_{step}_{ptor}')
                labels_slct.append(f'criterion_{step}_{ptor}')
                labels_slct.append(f'weight_{step}_{ptor}')
                labels_slct.append(f'time_{step}_{ptor}')
                labels_slct.append(f'x_min_{step}_{ptor}')
                labels_slct.append(f'x_max_{step}_{ptor}')
                labels_slct.append(f'y_min_{step}_{ptor}')
                labels_slct.append(f'y_max_{step}_{ptor}')

        # Extract values
        for filename in self.files:
            results = ParametersArray(filename)
            results.load()
            data_slct = []
            for step, ptors in enumerate(self.struct):
                data_slct.append(results.get_anbs(step))
                for ptor in range(ptors):
                    if ptor >= results.struct[step]:
                        continue
                    data_slct.append(results.get_variable_and_level(step, ptor))
                    data_slct.append(results.get_criterion(step, ptor))
                    data_slct.append(results.get_weight(step, ptor))
                    data_slct.append(results.get_time(step, ptor))
                    data_slct.append(results.get_xmin(step, ptor))
                    data_slct.append(results.get_xmax(step, ptor))
                    data_slct.append(results.get_ymin(step, ptor))
                    data_slct.append(results.get_ymax(step, ptor))

            vals = [int(results.get_station()), results.get_valid_score(),
                    filename] + data_slct
            data.append(vals)

        labels = ['station', 'score', 'file'] + labels_slct
        self.data = pd.DataFrame(data, columns=labels)
        self.data.sort_values(by=['station', 'file'], inplace=True)

    def _list_variables(self):
        for step, ptors in enumerate(self.struct):
            for ptor in range(ptors):
                variables = self.data[f'var_{step}_{ptor}']
                if self.filter_min_weight > 0:
                    weight = self.data[f'weight_{step}_{ptor}']
                    variables = variables[weight >= self.filter_min_weight]
                if len(self.vars) == 0:
                    self.vars = variables
                else:
                    self.vars = pd.concat([self.vars, variables])

        self.vars = self.vars.drop_duplicates()
        self.vars = self.vars.sort_values(ascending=False)
        if self.use_vars_default:
            old_vars = list(self.vars)
            new_vars = []
            for var in self.vars_default:
                if var in old_vars:
                    new_vars.append(var)
                    old_vars.remove(var)
            if len(old_vars) > 0:
                new_vars += old_vars
            new_vars.reverse()
            self.vars = pd.Series(new_vars)
        self.vars = self.vars.reset_index(drop=True)

        if len(self.vars) < self.variables_importance_nb:
            self.variables_importance_nb = len(self.vars)

    def _list_stations(self):
        self.stations = self.data['station']
        self.stations = self.stations.drop_duplicates()
        self.stations = self.stations.sort_values(ascending=False)
        self.stations = self.stations.reset_index(drop=True)

    def _add_variable_index(self):
        for step, ptors in enumerate(self.struct):
            for ptor in range(ptors):
                label = f'var_index_{step}_{ptor}'
                self.data[label] = None
                for idx, _ in enumerate(self.vars):
                    self.data.loc[
                        self.data[f'var_{step}_{ptor}'] == self.vars[idx], label] = idx

    def _set_criteria_color(self):
        for step, ptors in enumerate(self.struct):
            for ptor in range(ptors):
                label = f'crit_color_{step}_{ptor}'
                self.data[label] = None
                for icrit, crit in enumerate(self.crit):
                    indexes = self.data.loc[
                        self.data[f'criterion_{step}_{ptor}'] == crit].index
                    for index in indexes:
                        self.data[label].iloc[index] = self.colors[icrit]

    def _drop_bad_scores(self):
        for station in self.stations:
            indexes = self.data.loc[self.data['station'] == station].index
            scores = self.data['score'].loc[indexes]
            min_score = min(scores)
            sizes = self.marker_size_max * (
                    (min_score * (1 + self.marker_size_range)) - scores) / (
                            min_score * self.marker_size_range)
            drop_rows = indexes[sizes < 1]
            self.data.drop(index=drop_rows, inplace=True)
            self.data.reset_index(drop=True, inplace=True)

    def _set_marker_size(self):
        if self.marker_size_on_weight:
            max_weight = 0.2
            min_weight = 0.01
            for step, ptors in enumerate(self.struct):
                for ptor in range(ptors):
                    label = f'marker_size_{step}_{ptor}'
                    self.data[label] = None
                    sizes = []
                    for weight in self.data[f'weight_{step}_{ptor}']:
                        size = self.marker_size_max * (weight - min_weight) / (
                                    max_weight - min_weight)
                        if weight == 0:
                            size = 0
                        elif size < 1:
                            size = 1
                        elif size > self.marker_size_max:
                            size = self.marker_size_max
                        sizes.append(size)
                    self.data[label] = sizes
        else:
            self.data['marker_size'] = 0
            for station in self.stations:
                indexes = self.data.loc[self.data['station'] == station].index
                scores = self.data['score'].loc[indexes]
                min_score = min(scores)
                sizes = self.marker_size_max * (
                        (min_score * (1 + self.marker_size_range)) - scores) / (
                                min_score * self.marker_size_range)
                sizes[sizes < 1] = 1
                self.data.loc[indexes, 'marker_size'] = sizes

    def _make_scatter_plot(self, show_criteria=True):
        fig_height = 0.66 + float(len(self.vars)) * 3.7/25.0
        self.fig = plt.figure(figsize=(10, fig_height))
        plt.grid(axis='y', alpha=0.2)
        plt.xlim(0, len(self.data) + 1)
        plt.ylim(-1, len(self.vars))

        do_paint = False
        xticks = []
        for station in self.stations:
            indexes = self.data.loc[self.data['station'] == station].index
            index_min = min(indexes) + 1
            index_max = max(indexes) + 1
            xticks.append((index_min + index_max) / 2.0)
            if do_paint:
                plt.axvspan(index_min - 0.5, index_max + 0.5, facecolor='k', alpha=0.08)
            do_paint = not do_paint

        for step, ptors in enumerate(self.struct):
            marker = self.markers[step]
            for ptor in range(ptors):
                x = np.arange(1, len(self.data) + 1)
                variables = self.data[f'var_index_{step}_{ptor}']
                facecolors = None
                edgecolors = 'black'
                if show_criteria:
                    facecolors = self.data[f'crit_color_{step}_{ptor}']
                    edgecolors = self.data[f'crit_color_{step}_{ptor}']
                if self.marker_size_on_weight:
                    sizes = self.data[f'marker_size_{step}_{ptor}']
                    facecolors = 'none'
                else:
                    sizes = self.data['marker_size']

                nodata = variables.isnull()
                if nodata.any():
                    indices = variables[nodata].index.tolist()
                    variables = variables.drop(indices)
                    edgecolors = edgecolors.drop(indices)
                    sizes = sizes.drop(indices)
                    x = np.delete(x, indices)

                plt.scatter(x, variables, marker=marker, facecolors=facecolors,
                            edgecolors=edgecolors, s=sizes,
                            alpha=self.marker_alpha, zorder=10)
        plt.xticks(xticks, self.stations)
        plt.yticks(self.vars.index.tolist(), self.vars)
        plt.tick_params(axis='both', which='both', bottom=False, top=False,
                        labelbottom=True, left=False, right=False, labelleft=True)

        if show_criteria:
            patches = []
            for idx, criteria in enumerate(self.crit):
                patches.append(mpatches.Patch(color=self.colors[idx], label=criteria,
                                              alpha=self.marker_alpha))
            plt.legend(handles=patches, title="Criteria", loc="center left",
                       bbox_to_anchor=(1, 0, 0.5, 1), frameon=False)

        self.fig.tight_layout()

    def _make_syntheses_plot(self, show_criteria=True, target_lons=None,
                             target_lats=None, sort_by_selection=True):
        cmap = plt.get_cmap('viridis')

        fig_height = 2 + 2 * self.variables_importance_nb * 3.7/25.0
        self.fig, axs = plt.subplots(2, 2, figsize=(10, fig_height))
        axs[0, 1].grid(axis='y', alpha=0.2)
        axs[1, 1].grid(axis='y', alpha=0.2)
        axs[0, 0].set_ylim(0, self.variables_importance_nb + 1)
        axs[0, 1].set_ylim(0, self.variables_importance_nb + 1)
        axs[1, 0].set_ylim(0, self.variables_importance_nb + 1)
        axs[1, 1].set_ylim(0, self.variables_importance_nb + 1)

        self.legend, legend_axs = plt.subplots(1, 2, figsize=(10, 0.5))
        legend_axs[0].axis('off')
        legend_axs[1].axis('off')

        vars_weights = [[] for i in range(len(self.vars))]
        vars_colors = [[] for i in range(len(self.vars))]
        vars_times = [[] for i in range(len(self.vars))]
        vars_xmins = [[] for i in range(len(self.vars))]
        vars_xmaxs = [[] for i in range(len(self.vars))]
        vars_ymins = [[] for i in range(len(self.vars))]
        vars_ymaxs = [[] for i in range(len(self.vars))]

        for step, ptors in enumerate(self.struct):
            for ptor in range(ptors):
                variables = self.data[f'var_index_{step}_{ptor}']
                weights = self.data[f'weight_{step}_{ptor}']
                crit_color = cmap(0.5)
                if show_criteria:
                    crit_color = self.data[f'crit_color_{step}_{ptor}']
                times = self.data[f'time_{step}_{ptor}']
                xmins = self.data[f'x_min_{step}_{ptor}']
                xmaxs = self.data[f'x_max_{step}_{ptor}']
                ymins = self.data[f'y_min_{step}_{ptor}']
                ymaxs = self.data[f'y_max_{step}_{ptor}']

                nodata = variables.isnull()
                if nodata.any():
                    indices = variables[nodata].index.tolist()
                    variables = variables.drop(indices)
                    weights = weights.drop(indices)
                    crit_color = crit_color.drop(indices)
                    times = times.drop(indices)
                    xmins = xmins.drop(indices)
                    xmaxs = xmaxs.drop(indices)
                    ymins = ymins.drop(indices)
                    ymaxs = ymaxs.drop(indices)

                for i in range(len(variables)):
                    vars_weights[variables.iloc[i]].append(weights.iloc[i])
                    var_color = cmap(0.5)
                    if show_criteria:
                        var_color = crit_color.iloc[i]
                    vars_colors[variables.iloc[i]].append(var_color)
                    vars_times[variables.iloc[i]].append(times.iloc[i])
                    vars_xmins[variables.iloc[i]].append(xmins.iloc[i])
                    vars_xmaxs[variables.iloc[i]].append(xmaxs.iloc[i])
                    vars_ymins[variables.iloc[i]].append(ymins.iloc[i])
                    vars_ymaxs[variables.iloc[i]].append(ymaxs.iloc[i])

        sums = []
        counts = []
        for weights in vars_weights:
            sums.append(np.sum(weights))
            counts.append(len(weights))

        if sort_by_selection:
            counts, sums, vars_weights, vars_colors, vars_times, vars_xmins, \
                vars_xmaxs, vars_ymins, vars_ymaxs, vars = (
                    list(t) for t in zip(*sorted(
                        zip(counts, sums, vars_weights, vars_colors,
                            vars_times, vars_xmins, vars_xmaxs, vars_ymins,
                            vars_ymaxs, self.vars.tolist()),
                        reverse=True)))
        else:
            vars = self.vars.tolist()
            vars.reverse()
            counts.reverse()
            sums.reverse()
            vars_weights.reverse()
            vars_colors.reverse()
            vars_times.reverse()
            vars_xmins.reverse()
            vars_xmaxs.reverse()
            vars_ymins.reverse()
            vars_ymaxs.reverse()

        y = range(1, self.variables_importance_nb + 1)

        # Longitude / longitude colors
        color1 = 'tab:blue'
        color2 = 'tab:brown'
        flierprops1 = dict(marker='.', markeredgecolor=color1)
        flierprops2 = dict(marker='.', markeredgecolor=color2)

        # Print variables with criteria (color)
        count_colors = [[] for i in range(self.variables_importance_nb)]
        for i in range(self.variables_importance_nb):
            colors = vars_colors[i]
            count_col = Counter(colors)
            count_col = sorted(count_col.items(), key=lambda item: item[1],
                               reverse=True)
            count_colors[i] = count_col

        # Simple version: ax1.barh(y, counts[0:self.variables_importance_nb])
        for i in range(self.variables_importance_nb):
            x_start = 0
            for count_col in count_colors[i]:
                x_width = count_col[1]
                color = count_col[0]
                axs[0, 0].barh(i+1, x_width, left=x_start, color=color)
                x_start += x_width

        if show_criteria:
            patches = []
            for idx, criteria in enumerate(self.crit):
                patches.append(mpatches.Patch(color=self.colors[idx], label=criteria))
            axs[0, 0].legend(handles=patches, title="Criteria", loc="lower right",
                             frameon=False)

        # Print weights
        #  axs[0, 1].boxplot(vars_weights[0:self.variables_importance_nb], vert=False)

        # Print longitude
        if target_lons:
            axs[0, 1].axvspan(target_lons[0], target_lons[1],
                              facecolor='gray', alpha=0.3)
        box1 = axs[0, 1].boxplot(vars_xmins[0:self.variables_importance_nb], vert=False,
                                 flierprops=flierprops1)
        box2 = axs[0, 1].boxplot(vars_xmaxs[0:self.variables_importance_nb], vert=False,
                                 flierprops=flierprops2)
        for element, line_list in box1.items():
            if element == 'medians':
                continue
            for line in line_list:
                line.set_color(color1)
        for element, line_list in box2.items():
            if element == 'medians':
                continue
            for line in line_list:
                line.set_color(color2)

        # Print time
        count_times = [[] for i in range(self.variables_importance_nb)]
        h_min = 99
        h_max = 0
        for i in range(self.variables_importance_nb):
            times = vars_times[i]
            count_time = Counter(times)
            count_time = sorted(count_time.items(), key=lambda item: item[0])
            count_time_sum = np.sum(count_time, axis=0)
            count_time = [(x[0], 100 * x[1] / count_time_sum[1]) for x in count_time]
            count_times[i] = count_time
            h_min = min(float(np.min(count_time, axis=0)[0]), h_min)
            h_max = max(float(np.max(count_time, axis=0)[0]), h_max)

        for i in range(self.variables_importance_nb):
            x_start = 0
            for count_time in count_times[i]:
                x_width = count_time[1]
                time = count_time[0]
                color = cmap((time - h_min) / (h_max - h_min))
                axs[1, 0].barh(i+1, x_width, left=x_start, color=color)
                x_start += x_width

        patches = []
        for hr in range(int(h_min), int(h_max) + self.time_step, self.time_step):
            patches.append(mpatches.Patch(color=cmap((hr - h_min) / (h_max - h_min)),
                                          label=f'{hr}'))
        legend_axs[0].legend(handles=patches, bbox_to_anchor=(0.5, 0),
                             ncol=6, loc='lower center', frameon=False)

        # Print latitudes
        if target_lats:
            axs[1, 1].axvspan(target_lats[0], target_lats[1],
                              facecolor='gray', alpha=0.3)
        box1 = axs[1, 1].boxplot(vars_ymins[0:self.variables_importance_nb], vert=False,
                                 flierprops=flierprops1)
        box2 = axs[1, 1].boxplot(vars_ymaxs[0:self.variables_importance_nb], vert=False,
                                 flierprops=flierprops2)
        for element, line_list in box1.items():
            if element == 'medians':
                continue
            for line in line_list:
                line.set_color(color1)
        for element, line_list in box2.items():
            if element == 'medians':
                continue
            for line in line_list:
                line.set_color(color2)

        min_line = mlines.Line2D([], [], color=color1, label='min')
        max_line = mlines.Line2D([], [], color=color2, label='max')
        target = mpatches.Patch(color='gray', alpha=0.3, label='Target')
        patches.append(mpatches.Patch(color1, label='min'))
        patches.append(mpatches.Patch(color2, label='max'))
        legend_axs[1].legend(handles=[min_line, max_line, target],
                             bbox_to_anchor=(0.5, 0), ncol=3,
                             loc='lower center', frameon=False)

        y_ticks = vars[0:self.variables_importance_nb]
        axs[0, 0].set_yticks(y)
        axs[0, 0].set_yticklabels(y_ticks)
        axs[0, 0].invert_yaxis()
        axs[0, 1].set_yticks(y)
        axs[0, 1].set_yticklabels(y_ticks)
        axs[0, 1].invert_yaxis()
        axs[1, 0].set_yticks(y)
        axs[1, 0].set_yticklabels(y_ticks)
        axs[1, 0].invert_yaxis()
        axs[1, 1].set_yticks(y)
        axs[1, 1].set_yticklabels(y_ticks)
        axs[1, 1].invert_yaxis()

        axs[0, 0].set_xlabel('Number of selections')
        axs[0, 1].set_xlabel('Longitude min/max')
        axs[1, 0].set_xlabel('Percentage of selected hours (%)')
        axs[1, 1].set_xlabel('Latitude min/max')

        self.fig.tight_layout()

    def _print(self, filename):
        self.fig.savefig(os.path.join(self.output_path, filename + '.pdf'))
        self.fig.savefig(os.path.join(self.output_path, filename + '.png'), dpi=300)
        if self.legend:
            self.legend.savefig(os.path.join(
                self.output_path, filename + '_legend.pdf'))
            self.legend.savefig(os.path.join(
                self.output_path, filename + '_legend.png'), dpi=300)

        plt.close(self.fig)



from .cleanup import (
    cleanup_duplicate_ini_files,
    cleanup_empty_log_files,
    cleanup_old_generations,
)
from .compress import compress_log_files, compress_optimization_outputs
from .mjd import as_mjd, date_as_mjd, days_to_hours_mins, jd_to_date, mjd_to_datetime

__all__ = ('cleanup_empty_log_files', 'cleanup_duplicate_ini_files',
           'cleanup_old_generations', 'compress_optimization_outputs',
           'compress_log_files', 'as_mjd', 'date_as_mjd', 'jd_to_date',
           'days_to_hours_mins', 'mjd_to_datetime')



import tempfile

from atmoswing_toolbox.datasets import GenericDataset


def test_module_import():
    with tempfile.TemporaryDirectory() as tmp_dir:
        GenericDataset(directory=tmp_dir, var_name=None, ref_data=None)



import os

import numpy as np
from netCDF4 import Dataset


class Forecast:
    """Extract AtmoSwing Forecaster outputs"""

    def __init__(self, file_path=''):
        self.file_path = file_path
        self.version = 0
        self.target_dates = None
        self.analog_criteria = None
        self.analog_dates = None
        self.analog_values = None
        self.analogs_nb = None
        self.stations_names = None
        self.stations_ids = None
        self.stations_heights = None
        self.stations_x_coords = None
        self.stations_y_coords = None
        self.reference_axis = None
        self.reference_values = None
        self._opened = False
        self._file = None
        self._use_analogs_nb_t0 = False

    def open(self):
        if not os.path.isfile(self.file_path):
            print(f'File {self.file_path} not found')
            return False

        # Read content
        self._file = Dataset(self.file_path, 'r')

        # Get version
        self.version = self._file.version_major + self._file.version_minor / 10

        # Extract variables
        if self.version <= 1.1:
            self.target_dates = self._file.variables['targetdates']
            self.analog_criteria = self._file.variables['analogscriteria']
            self.analog_dates = self._file.variables['analogsdates']
            self.analog_values = self._file.variables['analogsvaluesgross']
            self.analogs_nb = self._file.variables['analogsnb']
            self.stations_names = self._file.variables['stationsnames']
            self.stations_ids = self._file.variables['stationsids']
            self.stations_heights = self._file.variables['stationsheights']
            self.stations_x_coords = self._file.variables['loccoordu']
            self.stations_y_coords = self._file.variables['loccoordv']
            self.reference_axis = self._file.variables['returnperiods']
            self.reference_values = self._file.variables[
                'dailyprecipitationsforreturnperiods']
        elif self.version <= 1.2:
            self.target_dates = self._file.variables['target_dates']
            self.analog_criteria = self._file.variables['analogs_criteria']
            self.analog_dates = self._file.variables['analogs_dates']
            self.analog_values = self._file.variables['analogs_values_gross']
            self.analogs_nb = self._file.variables['analogs_nb']
            self.stations_names = self._file.variables['stations_names']
            self.stations_ids = self._file.variables['stations_ids']
            self.stations_heights = self._file.variables['stations_heights']
            self.stations_x_coords = self._file.variables['loc_coord_u']
            self.stations_y_coords = self._file.variables['loc_coord_v']
            self.reference_axis = self._file.variables['reference_axis']
            self.reference_values = self._file.variables['reference_values']
        else:
            self.target_dates = self._file.variables['target_dates']
            self.analog_criteria = self._file.variables['analog_criteria']
            self.analog_dates = self._file.variables['analog_dates']
            self.analog_values = self._file.variables['analog_values']
            self.analogs_nb = self._file.variables['analogs_nb']
            self.stations_names = self._file.variables['station_names']
            self.stations_ids = self._file.variables['station_ids']
            self.stations_heights = self._file.variables['station_heights']
            self.stations_x_coords = self._file.variables['station_x_coords']
            self.stations_y_coords = self._file.variables['station_y_coords']
            self.reference_axis = self._file.variables['reference_axis']
            self.reference_values = self._file.variables['reference_values']

        self._opened = True

        return True

    def close(self):
        self._check_opened()
        self._file.close()

    def _check_opened(self):
        if not self._opened:
            raise Exception(f'File {self.file_path} not opened')

    def use_analogs_nb_t0(self):
        self._use_analogs_nb_t0 = True

    def get_analogs_nb(self, time_lapse):
        self._check_opened()
        if self._use_analogs_nb_t0:
            return int(np.array(self.analogs_nb[:])[0])
        else:
            return int(np.array(self.analogs_nb[:])[time_lapse])

    def _get_station_index(self, station_id):
        self._check_opened()
        stations_ids = np.array(self.stations_ids[:])
        i_stat = np.where(stations_ids == station_id)
        if i_stat[0].size == 0:
            raise Exception(f'Station id {station_id} not found')

        return i_stat[0]

    def get_station_name(self, station_id):
        self._check_opened()
        i_stat = self._get_station_index(station_id)

        return self.stations_names[i_stat][0]

    def get_reference(self, station_id):
        self._check_opened()
        i_stat = self._get_station_index(station_id)

        ref_axis = np.array(self.reference_axis)
        all_ref_values = np.array(self.reference_values[:, :])
        if self.version < 1.2:
            curr_shape = all_ref_values.shape
            new_shape = curr_shape[::-1]
            all_ref_values = np.reshape(all_ref_values, new_shape)
            all_ref_values = np.transpose(all_ref_values)
            ref_values = all_ref_values[:, i_stat]
        else:
            ref_values = all_ref_values[i_stat, :][0]

        return {'axis': ref_axis, 'values': ref_values}

    def get_analog_values(self, station_id, time_lapse):
        self._check_opened()
        i_stat = self._get_station_index(station_id)

        analog_values = np.array(self.analog_values[:, :])
        if self.version < 1.2:
            curr_shape = analog_values.shape
            new_shape = curr_shape[::-1]
            analog_values = np.reshape(analog_values, new_shape)
            analog_values = np.transpose(analog_values)

        nb_analogs = self.get_analogs_nb(time_lapse)

        col_start = 0
        for i_lapse in np.arange(time_lapse):
            col_start = col_start + np.array(self.analogs_nb[:])[i_lapse]
        col_end = col_start + nb_analogs

        return analog_values[i_stat, col_start:col_end]



import math
import os

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np


class ForecastTimeSeriesPlot:
    """Base class to create a time serie of forecast"""

    def __init__(self, output_path=''):
        self.fig = None
        self.ax = None
        self.output_path = output_path
        self.output_name = 'plot'
        self.dates = None
        self.obs_values = None
        self.analogs_values = None
        self.start = 0
        self.end = 0
        self.ref_periods = np.array([])
        self.ref_precip = np.array([])
        self.max_val = 0
        self._show_legend = True
        self._show_dots = False

    def show(self):
        plt.ion()
        self._make_figure()
        plt.show()

    def print(self):
        if not self.output_path:
            raise Exception('Output path not provided')
        plt.ioff()
        self._make_figure()
        self._print()

    def _make_figure(self):
        plt.rcParams.update({'font.size': 15})
        self.fig = plt.figure(figsize=(18, 4.5))
        self.ax = self.fig.add_subplot(111)
        self.build()

    def _print(self):
        self.fig.savefig(os.path.join(self.output_path, self.output_name + '.pdf'))
        self.fig.savefig(os.path.join(self.output_path, self.output_name + '.png'),
                         dpi=300)
        plt.close(self.fig)

    def set_output_name(self, name):
        self.output_name = name

    def limit_period(self, start, end):
        # Using datetime: ex: datetime.date(1987, 7, 4)
        self.start = start
        self.end = end

    def set_reference_levels(self, ref_periods, ref_precip):
        self.ref_periods = ref_periods
        self.ref_precip = ref_precip

    def hide_legend(self):
        self._show_legend = False

    def show_all_dots(self):
        self._show_dots = True

    def set_data(self, dates, obs_values, analogs_values):
        self.dates = dates
        self.obs_values = obs_values
        self.analogs_values = analogs_values

    def set_max_val(self, value):
        self.max_val = value

    def build(self):
        # Split data if we focus on a period
        if self.start and self.end:
            start_array = np.searchsorted(self.dates, self.start.toordinal() - 678576)
            end_array = np.searchsorted(self.dates, self.end.toordinal() - 678575)
            self.analogs_values = self.analogs_values[start_array:end_array, :]
            self.dates = self.dates[start_array:end_array]
            self.obs_values = self.obs_values[start_array:end_array]

        # Convert time for plotting
        plot_dates = self.dates + 678575 + 1

        # Extract quantiles
        q030 = np.nanmean(np.percentile(self.analogs_values, 30, axis=2), axis=0)
        q060 = np.nanmean(np.percentile(self.analogs_values, 60, axis=2), axis=0)
        q090_all = np.percentile(self.analogs_values, 90, axis=2)
        q090 = np.nanmean(q090_all, axis=0)
        q100 = np.nanmean(np.percentile(self.analogs_values, 100, axis=2), axis=0)

        # Plot the reference values
        plot_obs = self.obs_values
        plot_obs[np.isnan(q100)] = np.nan
        self.ax.plot(plot_dates, plot_obs, '-', linewidth=2, color='r',
                     label="observations")

        # Plot areas and lines
        self.ax.fill_between(plot_dates, q030, q060, facecolor=[0.6, 0.6, 0.6],
                             edgecolor='None')
        self.ax.fill_between(plot_dates, q060, q090, facecolor=[0.6, 0.6, 0.6],
                             edgecolor='None')
        self.ax.plot(plot_dates, q030, '--', linewidth=1, color='k',
                     label=r"mean quantile 30\%")
        self.ax.plot(plot_dates, q060, '-', linewidth=1, color='k',
                     label=r"mean quantile 60\%")
        self.ax.plot(plot_dates, q090_all[0], ':', linewidth=1, color='k',
                     label=r"single quantile 90\%")
        self.ax.plot(plot_dates, q090_all[1], ':', linewidth=1, color='k')
        self.ax.plot(plot_dates, q090_all[2], ':', linewidth=1, color='k')
        self.ax.plot(plot_dates, q090_all[3], ':', linewidth=1, color='k')
        self.ax.plot(plot_dates, q090, '--', linewidth=1, color='k',
                     label=r"mean quantile 90\%")
        self.ax.plot(plot_dates, q100, 'x', markersize=3, color='0.5',
                     label="mean maximum")

        # Gray nan areas
        for idx, val in enumerate(q100):
            if math.isnan(val):
                self.ax.axvspan(plot_dates[max(idx - 1, 0)],
                                plot_dates[min(idx + 1, plot_dates.size - 1)],
                                facecolor=[0.9, 0.9, 0.9], edgecolor='None')

        # Plot the dots
        if self._show_dots:
            for i in range(self.analogs_values.shape[0]):
                self.ax.plot(plot_dates, self.analogs_values[:, i], 'ko')

        # Format the ticks
        if self.start and self.end:
            days = mdates.DayLocator()
            some_days = mdates.DayLocator(interval=5)
            days_fmt = mdates.DateFormatter('%d.%m.%Y')
            self.ax.xaxis.set_major_locator(some_days)
            self.ax.xaxis.set_minor_locator(days)
            self.ax.xaxis.set_major_formatter(days_fmt)
        else:
            self.ax.xaxis.set_major_locator(mdates.MonthLocator())
            self.ax.xaxis.set_minor_locator(mdates.MonthLocator(bymonthday=15))
            self.ax.xaxis.set_major_formatter(ticker.NullFormatter())
            self.ax.xaxis.set_minor_formatter(mdates.DateFormatter('%b'))

            for tick in self.ax.xaxis.get_minor_ticks():
                tick.tick1line.set_markersize(0)
                tick.tick2line.set_markersize(0)
                tick.label1.set_horizontalalignment('center')

            i_mid = len(plot_dates) // 2
            date = plot_dates[i_mid]
            if np.isnan(date):
                date = plot_dates[0]
            if np.isnan(date):
                date = plot_dates[-1]
            y = mdates.num2date(date).year
            self.ax.set_xlabel(str(y))

        # Draw grid and set axis label
        self.ax.grid(True)
        plt.ylabel('Precipitation (mm/d)')

        # Draw line return periods P10
        if self.ref_periods.size == 0 and self.ref_precip.size == 0:
            p10_index = np.searchsorted(self.ref_periods, 10)
            plt.axhline(y=self.ref_precip[p10_index], linewidth=2, color='r',
                        label="return period of 10 years")

        # Set correct limits
        plt.ylim(bottom=0)
        if self.max_val > 0:
            plt.ylim(top=self.max_val)

        if self.start and self.end:
            self.ax.set_xlim(self.start, self.end)
        else:
            plt.xlim(plot_dates[0], plot_dates[-1])

        # Legends
        if self._show_legend:
            handles, labels = self.ax.get_legend_handles_labels()
            self.ax.legend(handles, labels, loc='upper left')

        self.fig.tight_layout()



import pandas as pd


class ParametersArray:
    """Parsing of the parameters resulting from the optimizer"""

    def __init__(self, path):
        self.path = path
        self.id = 0
        self.score = ''
        self.struct = []
        self.data = []

    def load(self):
        file_struct = self._parse_headers()
        # Check for duplicates
        seen = set()
        dupes = [x for x in file_struct[1] if x in seen or seen.add(x)]
        if len(dupes) > 0:
            raise Exception(f'Duplicates found in the column names: {dupes}')
        self.data = pd.read_csv(self.path, sep='\t', skiprows=1, usecols=file_struct[0],
                                names=file_struct[1])
        self._remove_failed()

    def load_scores_only(self):
        file_struct = self._parse_headers_scores_only()
        self.data = pd.read_csv(self.path, sep='\t', skiprows=1, usecols=file_struct[0],
                                names=file_struct[1])
        self._remove_failed()

    def _parse_headers(self):
        fid = open(self.path)
        fid.readline()
        line = fid.readline().split('\t')
        self.id = line[1]

        headers = []
        cols = []
        step = 0
        ptor = 0
        for idx, chars in enumerate(line):
            append = False
            if chars[0:10] == '|||| Step(':
                step = int(chars[10:-1])
                self.struct.append(0)
            elif chars[0:8] == '|| Ptor(':
                ptor = int(chars[8:-1])
                self.struct[step] = ptor + 1
            elif chars == 'Anb':
                label = f'Anb_{step}'
                headers.append(label)
                cols.append(idx + 1)
            elif chars == 'Level':
                append = True
                label = f'Variable_{step}_{ptor}'
                if label in headers:
                    label += 'b'
                headers.append(label)
                cols.append(idx-1)
            elif chars == 'Time':
                append = True
            elif chars == 'xMin':
                append = True
            elif chars == 'xPtsNb':
                append = True
            elif chars == 'xStep':
                append = True
            elif chars == 'yMin':
                append = True
            elif chars == 'yPtsNb':
                append = True
            elif chars == 'yStep':
                append = True
            elif chars == 'Weight':
                append = True
            elif chars == 'Criteria':
                append = True
            elif chars == 'Calib':
                headers.append(chars)
                cols.append(idx + 1)
            elif chars == 'Valid':
                headers.append(chars)
                cols.append(idx + 1)
            elif "Score" in chars:
                self.score = line[idx + 1]

            if append:
                label = f'{chars}_{step}_{ptor}'
                if label in headers:
                    label += 'b'
                headers.append(label)
                cols.append(idx + 1)

        fid.close()

        return [cols, headers]

    def _parse_headers_scores_only(self):
        fid = open(self.path)
        fid.readline()
        line = fid.readline().split('\t')
        self.id = line[1]

        headers = []
        cols = []
        for idx, chars in enumerate(line):
            if chars == 'Calib':
                headers.append(chars)
                cols.append(idx + 1)
            elif chars == 'Valid':
                headers.append(chars)
                cols.append(idx + 1)
            elif "Score" in chars:
                self.score = line[idx + 1]

        fid.close()

        return [cols, headers]

    def _remove_failed(self):
        if self.score == 'CRPS':
            # Remove 0s
            self.data.drop(self.data[self.data.Calib == 0].index, inplace=True)

    def get_station(self):
        return self.id

    def get_station_int(self):
        return int(self.id)

    def get_steps_nb(self):
        return len(self.struct)

    def get_ptors_nb(self, step):
        return self.struct[step]

    def get_anbs(self, step):
        return int(self.data[f'Anb_{step}'])

    def get_variables(self, step, ptor):
        return self.data[f'Variable_{step}_{ptor}']

    def get_variable(self, step, ptor, index=0):
        var = str(self.data[f'Variable_{step}_{ptor}'].iloc[index])
        # Remove prefixes
        if 'GenericNetcdf 100/' in var:
            var = var.replace('GenericNetcdf 100/', '')

        # Map levels
        if 'pressure/' in var:
            var = var.replace('pressure/', 'pl/')
        if 'press/' in var:
            var = var.replace('press/', 'pl/')
        if 'sff/' in var:
            var = var.replace('sff/', 'surf/')
        if 'sfa/' in var:
            var = var.replace('sfa/', 'surf/')

        # Check if composed variable
        if f'Variable_{step}_{ptor}b' in self.data:
            var2 = str(self.data[f'Variable_{step}_{ptor}b'].iloc[index])
            if 'pl/r' in var and 'tcw' in var2:
                return 'MI'
            else:
                return 'unknown'

        # Rename variables
        if 'hgt' in var:
            return 'Z'
        if var == 'z':
            return 'Z'
        elif 'pl/z' in var:
            return 'Z'
        elif 'pl/gpa' in var:
            return 'ZA'
        elif 'pl/w' in var:
            return 'W'
        elif 'pl/vvel' in var:
            return 'W'
        elif 'pl/vo' in var:  # Vorticity (relative) (ERA-int)
            return 'VO'
        elif 'pl/vpot' in var:  # Velocity potential (JRA55)
            return 'VPOT'
        elif 'pl/vgrd' in var:
            return 'V'
        elif 'pl/v' in var:
            return 'V'
        elif 'pl/u' in var:
            return 'U'
        elif 'pl/depr' in var:  # Dew-point depression (JRA55)
            return 'DEPR'
        elif 'surf/depr' in var:  # Dew-point depression (JRA55)
            return 'DEPR'
        elif 'pl/reld' in var:  # Relative divergence (JRA55)
            return 'RD'
        elif 'pl/d' in var:
            return 'D'
        elif 'pt/d' in var:
            return 'PT/D'
        elif 'pl/t' in var:
            return 'T'
        elif 'pl/pv' in var:
            return 'PV'
        elif 'pt/pv' in var:
            return 'PT/PV'
        elif 'pl/rh' in var:
            return 'RH'
        elif 'pl/r' in var:
            return 'RH'
        elif 'pv/z' in var:
            return 'PV/Z'
        elif 'pv/v' in var:
            return 'PV/V'
        elif 'pv/u' in var:
            return 'PV/U'
        elif 'pt/v' in var:
            return 'PT/V'
        elif 'pt/u' in var:
            return 'PT/U'
        elif 'pv/pres' in var:
            return 'PV/PRES'
        elif 'pt/pres' in var:
            return 'PT/PRES'
        elif 'pv/pt' in var:
            return 'PV/PT'
        elif 'pt/q' in var:
            return 'PT/SH'
        elif 'pl/q' in var:
            return 'SH'
        elif 'pt/mont' in var:
            return 'PT/MONT'
        elif 'surf/strd' in var:
            return 'STRD'
        elif 'single/strd' in var:
            return 'STRD'
        elif 'surf/str' in var:
            return 'STR'
        elif 'single/str' in var:
            return 'STR'
        elif 'surf/ssrd' in var:
            return 'SSRD'
        elif 'single/ssrd' in var:
            return 'SSRD'
        elif 'surf/ssr' in var:
            return 'SSR'
        elif 'single/ssr' in var:
            return 'SSR'
        elif 'surf/cape' in var:
            return 'CAPE'
        elif 'single/cape' in var:
            return 'CAPE'
        elif 'surf/ie' in var:
            return 'IE'
        elif 'surf/v10' in var:
            return 'V10m'
        elif 'single/v10' in var:
            return 'V10m'
        elif 'surf/u10' in var:
            return 'U10m'
        elif 'surf/msl' in var:
            return 'SLP'
        elif 'surf/rh' in var:
            return 'RH'
        elif 'surf/t2m' in var:
            return 'T2m'
        elif 'single/t2m' in var:
            return 'T2m'
        elif 'surf/sst' in var:
            return 'SST'
        elif 'surf/d2m' in var:
            return 'D2m'
        elif 'surf/sd' in var:
            return 'SD'
        elif 'surf/prmsl' in var:
            return 'SLP'
        elif 'single/tsr' in var:
            return 'TSR'
        elif 'single/ttr' in var:
            return 'TTR'
        elif 'single/sshf' in var:
            return 'SSHF'
        elif 'single/slhf' in var:
            return 'SLHF'
        elif 'msl/pres' in var:
            return 'SLP'
        elif 'single/lcc' in var:
            return 'LCC'
        elif 'pl/cc' in var:
            return 'CC'
        elif 'single/tcc' in var:
            return 'TCC'
        elif 'single/deg0l' in var:
            return 'DEG0L'
        elif 'plf/cwat' in var:
            return 'CWAT'
        elif 'ea/cwat' in var:
            return 'CWAT'
        elif 'tcf/cw' in var:
            return 'CWAT'
        elif 'tc/vwv' in var:
            return 'VWV'
        elif 'tcw' in var:
            return 'TCW'
        elif 'omega' in var:
            return 'W'
        return var

    def get_variable_and_level(self, step, ptor, index=0):
        var = self.get_variable(step, ptor, index)
        level = self.get_level(step, ptor, index)
        if level != 0:
            var += str(level)
        return var

    def get_levels(self, step, ptor):
        return self.data[f'Level_{step}_{ptor}']

    def get_level(self, step, ptor, index=0):
        return self.data[f'Level_{step}_{ptor}'].iloc[index]

    def get_times(self, step, ptor):
        return self.data[f'Time_{step}_{ptor}']

    def get_time(self, step, ptor, index=0):
        return self.data[f'Time_{step}_{ptor}'].iloc[index]

    def get_xmins(self, step, ptor):
        return self.data[f'xMin_{step}_{ptor}']

    def get_xmin(self, step, ptor, index=0):
        return self.data[f'xMin_{step}_{ptor}'].iloc[index]

    def get_xmaxs(self, step, ptor):
        return self.data[f'xMin_{step}_{ptor}'] + \
               (self.data[f'xPtsNb_{step}_{ptor}'] - 1) * self.data[
                   f'xStep_{step}_{ptor}']

    def get_xmax(self, step, ptor, index=0):
        return self.data[f'xMin_{step}_{ptor}'].iloc[index] + \
               (self.data[f'xPtsNb_{step}_{ptor}'].iloc[index] - 1) * \
               self.data[f'xStep_{step}_{ptor}'].iloc[index]

    def get_xptsnbs(self, step, ptor):
        return self.data[f'xPtsNb_{step}_{ptor}']

    def get_ymins(self, step, ptor):
        return self.data[f'yMin_{step}_{ptor}']

    def get_ymin(self, step, ptor, index=0):
        return self.data[f'yMin_{step}_{ptor}'].iloc[index]

    def get_ymaxs(self, step, ptor):
        return self.data[f'yMin_{step}_{ptor}'] + \
               (self.data[f'yPtsNb_{step}_{ptor}'] - 1) * self.data[
                   f'yStep_{step}_{ptor}']

    def get_ymax(self, step, ptor, index=0):
        return self.data[f'yMin_{step}_{ptor}'].iloc[index] + \
               (self.data[f'yPtsNb_{step}_{ptor}'].iloc[index] - 1) * \
               self.data[f'yStep_{step}_{ptor}'].iloc[index]

    def get_yptsnbs(self, step, ptor):
        return self.data[f'yPtsNb_{step}_{ptor}']

    def get_weight(self, step, ptor, index=0):
        return self.data[f'Weight_{step}_{ptor}'].iloc[index]

    def get_weights(self, step, ptor):
        return self.data[f'Weight_{step}_{ptor}']

    def get_criterion(self, step, ptor, index=0):
        criterion = self.data[f'Criteria_{step}_{ptor}'].iloc[index]
        if 'grads' in criterion:
            criterion = criterion[0:-5]
        return criterion

    def get_criteria(self, step, ptor):
        return self.data[f'Criteria_{step}_{ptor}']

    def get_calib_scores(self):
        return self.data['Calib']

    def get_calib_score(self, index=0):
        return self.data['Calib'].iloc[index]

    def get_valid_scores(self):
        return self.data['Valid']

    def get_valid_score(self, index=0):
        return self.data['Valid'].iloc[index]



import glob
import os

import dateutil.parser
import numpy as np

import atmoswing_toolbox as astb
from atmoswing_toolbox.utils import mjd

from .predictor_dataset import PredictorDataset


class GribDataset(PredictorDataset):
    """Extract Grib data"""

    def __init__(self, directory, file_pattern):
        super().__init__(directory, file_pattern)
        self.grib_version = 0
        self.param_code_1 = []
        self.param_code_2 = []
        self.param_code_3 = []

        if not astb.has_eccodes:
            raise ImportError("eccodes is required to do this.")

    def load(self):
        try:
            self._list()
            self._extract()
        except OSError as err:
            print(f"OS error: {err}")
        except Exception as e:
            print("Unexpected error:", e)
            raise

    def _list(self):
        if not os.path.isdir(self.directory):
            raise Exception(f'Directory {self.directory} not found')

        self._files = glob.glob(os.path.join(self.directory, self.file_pattern))

        if len(self._files) == 0:
            raise Exception(
                f'No file found as {os.path.join(self.directory, self.file_pattern)}')

        self._files.sort()

    def _extract(self):
        for file in self._files:
            if not os.path.isfile(file):
                raise Exception(f'File {file} not found')

            print('Reading ' + file)

            f = open(file, 'rb')

            while True:
                msgid = astb.eccodes.codes_new_from_file(
                    f, astb.eccodes.CODES_PRODUCT_GRIB)

                if msgid is None:
                    break

                if self.grib_version == 0:
                    self.grib_version = astb.eccodes.codes_get_long(
                        msgid, "editionNumber")

                self._extract_axes(msgid)
                self._extract_level(msgid)
                self._extract_time(msgid)
                self._extract_grib_code(msgid)

                data = astb.eccodes.codes_get_array(msgid, "values")
                n_lats = len(self.axis_lat)
                n_lons = len(self.axis_lon)
                if len(self.data) == 0:
                    self.data = data.reshape((1, 1, n_lats, n_lons))
                else:
                    if self.data.shape[2] != n_lats or self.data.shape[3] != n_lons:
                        raise Exception("Issues in the size of the arrays.")
                    self.data = np.append(self.data,
                                          data.reshape((1, 1, n_lats, n_lons)),
                                          axis=0)

                units = astb.eccodes.codes_get(msgid, "units")
                if self.data_units is None:
                    self.data_units = units
                if self.data_units != units:
                    raise Exception("Only 1 type of data per file supported so far.")

                astb.eccodes.codes_release(msgid)

            f.close()

    def _extract_axes(self, msgid):
        n_lats = astb.eccodes.codes_get_long(msgid, "Nj")
        n_lons = astb.eccodes.codes_get_long(msgid, "Ni")
        lat_start = astb.eccodes.codes_get_long(
            msgid, "latitudeOfFirstGridPointInDegrees")
        lon_start = astb.eccodes.codes_get_long(
            msgid, "longitudeOfFirstGridPointInDegrees")
        lat_end = astb.eccodes.codes_get_long(
            msgid, "latitudeOfLastGridPointInDegrees")
        lon_end = astb.eccodes.codes_get_long(
            msgid, "longitudeOfLastGridPointInDegrees")
        if lon_end < lon_start:
            lon_start -= 360

        axis_lat = np.linspace(lat_start, lat_end, n_lats)
        axis_lon = np.linspace(lon_start, lon_end, n_lons)

        if len(self.axis_lat) == 0:
            self.axis_lat = axis_lat
            self.axis_lon = axis_lon
        else:
            if not np.array_equal(self.axis_lat, axis_lat):
                raise Exception("Only 1 extent per file is supported so far.")
            if not np.array_equal(self.axis_lon, axis_lon):
                raise Exception("Only 1 extent per file is supported so far.")

    def _extract_level(self, msgid):
        level = astb.eccodes.codes_get_double(msgid, "level")
        type = astb.eccodes.codes_get(msgid, "typeOfLevel")
        if type == "isobaricInPa":
            level /= 100

        if len(self.axis_level) == 0:
            self.axis_level.append(level)
        else:
            if self.axis_level[0] != level:
                raise Exception("Only 1 level per file is supported so far.")

    def _extract_time(self, msgid):
        ref_date = dateutil.parser.parse(
            astb.eccodes.codes_get_string(msgid, "dataDate"))
        ref_date_mjd = mjd.as_mjd(ref_date.year, ref_date.month, ref_date.day)
        ref_time = astb.eccodes.codes_get(msgid, "dataTime")

        forecast_time = 0
        if self.grib_version == 2:
            forecast_time = astb.eccodes.codes_get_double(msgid, "forecastTime")
        elif self.grib_version == 1:
            forecast_time = astb.eccodes.codes_get_double(msgid, "endStep")

        time_unit = astb.eccodes.codes_get_long(msgid, "stepUnits")
        if time_unit == 0:
            forecast_time /= 1440
        elif time_unit == 1:
            forecast_time /= 24

        time = ref_date_mjd + ref_time / 24 + forecast_time

        self.axis_time.append(time)

    def _extract_grib_code(self, msgid):
        discipline = 0
        category = 0
        number = 0
        if self.grib_version == 2:
            discipline = astb.eccodes.codes_get_long(msgid, "discipline")
            category = astb.eccodes.codes_get_long(msgid, "parameterCategory")
            number = astb.eccodes.codes_get_long(msgid, "parameterNumber")
        elif self.grib_version == 1:
            category = astb.eccodes.codes_get_long(msgid, "table2Version")
            number = astb.eccodes.codes_get_long(msgid, "indicatorOfParameter")

        self.param_code_1.append(discipline)
        self.param_code_2.append(category)
        self.param_code_3.append(number)



import os

from netCDF4 import Dataset

NETCDF_3 = 0
""" NetCDF 3 format - larger but faster """
NETCDF_4 = 1
""" NetCDF 4 format - smaller but longer to read """


class GenericDataset:
    """Generic dataset"""

    def __init__(self, directory, var_name, ref_data):
        self.directory = directory
        self.var_name = var_name
        self.ref_data = ref_data
        if not os.path.exists(directory):
            os.makedirs(directory)

    def generate(self, format=NETCDF_3, file_name=None):
        shape = self.ref_data.data.shape  # time, level, lat, lon

        if file_name is None:
            file_name = self.var_name + '.nc'
        file_path = os.path.join(self.directory, file_name)
        if format == NETCDF_3:
            nc = Dataset(file_path, "w", format="NETCDF3_CLASSIC")
        else:
            nc = Dataset(file_path, "w", format="NETCDF4")

        # Global attributes
        nc.standardized = 0
        if self.ref_data.is_standardized:
            nc.standardization_mean = self.ref_data.mean
            nc.standardization_sd = self.ref_data.sd

        # Dimensions
        nc.createDimension("time", shape[0])
        nc.createDimension("level", shape[1])
        nc.createDimension("lat", shape[2])
        nc.createDimension("lon", shape[3])

        # Variables
        var_level = nc.createVariable("level", "f4", ("level",))
        var_lat = nc.createVariable("lat", "f4", ("lat",))
        var_lon = nc.createVariable("lon", "f4", ("lon",))
        if format == NETCDF_3:
            var_time = nc.createVariable("time", "f4", ("time",))
            var_data = nc.createVariable(self.var_name, "f4",
                                         ("time", "level", "lat", "lon",))
        else:
            var_time = nc.createVariable("time", "f4", ("time",),
                                         zlib=True, complevel=6,
                                         least_significant_digit=4)
            var_data = nc.createVariable(self.var_name, "f4",
                                         ("time", "level", "lat", "lon",),
                                         zlib=True, complevel=6)

        # Attributes
        var_time.setncattr('name', 'time')
        var_time.long_name = 'time'
        var_time.units = 'days since 1858-11-17 00:00:00.0 +0:00'
        var_time.calendar = 'standard'
        var_lat.setncattr('name', 'lat')
        var_lat.long_name = 'latitude'
        var_lat.units = 'degree_north'
        var_lon.setncattr('name', 'lon')
        var_lon.long_name = 'longitude'
        var_lon.units = 'degree_east'
        var_data.units = self.ref_data.data_units

        # Set data
        var_time[:] = self.ref_data.axis_time
        var_level[:] = self.ref_data.axis_level
        var_lat[:] = self.ref_data.axis_lat
        var_lon[:] = self.ref_data.axis_lon
        var_data[:] = self.ref_data.data

        nc.close()



from .forecast_timeseries import ForecastTimeSeriesPlot
from .gas_variables import GAsVariablesPlot
from .montecarlo import MonteCarloPlot

__all__ = ('ForecastTimeSeriesPlot', 'GAsVariablesPlot', 'MonteCarloPlot')



from setuptools import setup

with open('README.md') as f:
    readme = f.read()

setup(
    name="atmoswing-toolbox",
    version="1.3.3",
    author="Pascal Horton",
    author_email="pascal.horton@giub.unibe",
    description="Python tools for AtmoSwing",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=['atmoswing_toolbox', 'atmoswing_toolbox.datasets',
              'atmoswing_toolbox.parsers', 'atmoswing_toolbox.plots',
              'atmoswing_toolbox.utils'],
    package_dir={'atmoswing_toolbox': 'atmoswing_toolbox/datasets',
                 'atmoswing_toolbox.datasets': 'atmoswing_toolbox/datasets',
                 'atmoswing_toolbox.parsers': 'atmoswing_toolbox/parsers',
                 'atmoswing_toolbox.plots': 'atmoswing_toolbox/plots',
                 'atmoswing_toolbox.utils': 'atmoswing_toolbox/utils',
                 },
    zip_safe=False,
    extras_require={"test": ["pytest>=6.0"]},
    python_requires=">=3.8",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    readme="README.md",
    project_urls={
        "Source Code": "https://github.com/atmoswing/atmoswing-python-toolbox",
        "Bug Tracker": "https://github.com/atmoswing/atmoswing-python-toolbox/issues",
    },
    license="MIT",
)



import glob
import os

import dateutil.parser
import numpy as np

import atmoswing_toolbox as astb
from atmoswing_toolbox.utils import mjd

from .generic_dataset import GenericDataset
from .predictor_dataset import PredictorDataset


class NetcdfDataset(PredictorDataset):
    """Extract NetCDF data"""

    def __init__(self, directory, file_pattern, var_name):
        super().__init__(directory, file_pattern)
        self.var_name = var_name

        if not astb.has_netcdf:
            raise ImportError("netCDF4 is required to do this.")

    def load(self, spatial_stride=0):
        self._list()
        self._extract(spatial_stride)

    def create_generic_individual_files(self, directory, var_name, spatial_stride=0):
        self._list()
        for file in self._files:
            self._extract_file(file, spatial_stride)
            new_reanalysis = GenericDataset(directory=directory, var_name=var_name,
                                            ref_data=self)
            new_reanalysis.generate(file_name=os.path.basename(file))
            self._drop_data()

    def _list(self):
        if not os.path.isdir(self.directory):
            raise Exception(f'Directory {self.directory} not found')

        self._files = glob.glob(os.path.join(self.directory, self.file_pattern))

        if len(self._files) == 0:
            raise Exception(
                f'No file found as {os.path.join(self.directory, self.file_pattern)}')

        self._files.sort()

    def _extract(self, spatial_stride=0):
        for file in self._files:
            self._extract_file(file, spatial_stride)

    def _extract_file(self, file, stride=0):
        if not os.path.isfile(file):
            raise Exception(f'File {file} not found')

        print('Reading ' + file)
        nc = astb.Dataset(file, 'r')
        var = nc.variables[self.var_name]

        has_levels = len(var.dimensions) == 4

        if stride > 1:
            if has_levels:
                dat = nc.variables[self.var_name][:, :, 0::stride, 0::stride]
            else:
                dat = nc.variables[self.var_name][:, 0::stride, 0::stride]
            data = np.array(dat)
        else:
            data = np.array(var)

        time = np.array(nc.variables[var.dimensions[0]])
        time = self._convert_time(nc, var, time)

        if not has_levels:
            new_shape = (data.shape[0], 1, data.shape[1], data.shape[2])
            data = np.reshape(data, new_shape)

        if len(self.data) == 0:
            self.data = data
            self.data_units = nc.variables[self.var_name].units
            self.axis_time = time
            if has_levels:
                self.axis_level = np.array(nc.variables[var.dimensions[1]])
                self.axis_lat = np.array(nc.variables[var.dimensions[2]])
                self.axis_lon = np.array(nc.variables[var.dimensions[3]])
                if stride > 1:
                    self.axis_lat = np.array(
                        nc.variables[var.dimensions[2]][0::stride])
                    self.axis_lon = np.array(
                        nc.variables[var.dimensions[3]][0::stride])
            else:
                self.axis_level = [0]
                self.axis_lat = np.array(nc.variables[var.dimensions[1]])
                self.axis_lon = np.array(nc.variables[var.dimensions[2]])
                if stride > 1:
                    self.axis_lat = np.array(
                        nc.variables[var.dimensions[1]][0::stride])
                    self.axis_lon = np.array(
                        nc.variables[var.dimensions[2]][0::stride])
        else:
            self.data = np.append(self.data, data, axis=0)
            self.axis_time = np.append(self.axis_time, time, axis=0)

        nc.close()

    def _drop_data(self):
        self.data = []
        self.data_units = []
        self.axis_time = []
        self.axis_level = []
        self.axis_lat = []
        self.axis_lon = []

    @staticmethod
    def _convert_time(nc, var, time):
        time_units = nc.variables[var.dimensions[0]].units
        str_space = time_units.find(' ')
        time_step = time_units[0:str_space]
        str_since = time_units.find('since')
        str_date = time_units[str_since + 6:]
        str_space = str_date.find(' ')
        if str_space > 0:
            str_date = str_date[0:str_space]

        ref_date = dateutil.parser.parse(str_date)
        ref_date_mjd = mjd.as_mjd(ref_date.year, ref_date.month, ref_date.day)

        if time_step == "hours":
            time = time / 24

        time = time + ref_date_mjd

        return time



import datetime

import numpy as np
import pandas as pd


def as_mjd(year, month, day, hour=0, minute=0):
    date = datetime.datetime(year, month, day, hour, minute)
    return pd.DatetimeIndex([date]).to_julian_date() - 2400000.5


def date_as_mjd(date):
    return pd.DatetimeIndex(date).to_julian_date() - 2400000.5


def jd_to_date(jd):
    """
    Transform julian date numbers to year, month and day (array-based).
    From https://gist.github.com/jiffyclub/1294443
    """
    jd = jd + 0.5

    f, i = np.modf(jd)
    i = i.astype(int)

    a = np.trunc((i - 1867216.25) / 36524.25)
    b = np.zeros(len(jd))

    idx = tuple([i > 2299160])
    b[idx] = i[idx] + 1 + a[idx] - np.trunc(a[idx] / 4.)
    idx = tuple([i <= 2299160])
    b[idx] = i[idx]

    c = b + 1524
    d = np.trunc((c - 122.1) / 365.25)
    e = np.trunc(365.25 * d)
    g = np.trunc((c - e) / 30.6001)

    day = c - e + f - np.trunc(30.6001 * g)

    month = np.zeros(len(jd))
    month[g < 13.5] = g[g < 13.5] - 1
    month[g >= 13.5] = g[g >= 13.5] - 13
    month = month.astype(int)

    year = np.zeros(len(jd))
    year[month > 2.5] = d[month > 2.5] - 4716
    year[month <= 2.5] = d[month <= 2.5] - 4715
    year = year.astype(int)

    return year, month, day


def days_to_hours_mins(days):
    """Transform a number of days to hours and minutes"""
    hours = days * 24.
    hours, hour = np.modf(hours)

    minutes = hours * 60.
    _, minute = np.modf(minutes)

    return hour.astype(int), minute.astype(int)


def mjd_to_datetime(mjd):
    """Transform modified julian dates to datetime instances (array-based)."""
    jd = mjd + 2400000.5
    year, month, day = jd_to_date(jd)

    frac_days, day = np.modf(day)
    day = day.astype(int)

    hour, minute = days_to_hours_mins(frac_days)

    date = np.empty(len(mjd), dtype='datetime64[s]')

    for idx, _ in enumerate(year):
        date[idx] = datetime.datetime(year[idx], month[idx], day[idx],
                                      hour[idx], minute[idx], 0, 0)

    return date



import glob
import gzip
import os
import shutil
import zipfile


def compress_optimization_outputs(base_path):
    def pack_file(f):
        with open(f, 'rb') as f_in:
            with gzip.open(f + '.gz', 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.remove(f)
        print(f + ' compressed')

    for x in os.listdir(base_path):
        path = os.path.join(base_path, x, 'results')
        final = glob.glob(path + '/*best_individual.txt')
        if len(final) > 0:
            # If optimization ended, compress all files.
            files_generations = glob.glob(path + '/*generations.txt')
            files_operators = glob.glob(path + '/*operators.txt')
            for file in files_generations:
                pack_file(file)
            for file in files_operators:
                pack_file(file)
        else:
            # If not, compress all files but the last.
            files_generations = glob.glob(path + '/*generations.txt')
            files_generations.sort()
            files_operators = glob.glob(path + '/*operators.txt')
            files_operators.sort()
            for file in files_generations[:-1]:
                pack_file(file)
            for file in files_operators[:-1]:
                pack_file(file)


def compress_log_files(base_path):
    for x in os.listdir(base_path):
        path = os.path.join(base_path, x)
        print(path)
        log_file = glob.glob(path + '/AtmoSwingOptimizer*.log')
        if len(log_file) > 0:
            zip_obj = zipfile.ZipFile(path + '/log.zip', 'w', zipfile.ZIP_DEFLATED)
            for f in log_file:
                file_name = f.replace(path, '')
                file_name = file_name[1:]
                zip_obj.write(f, file_name)
                os.remove(f)
                print(file_name + ' compressed')
            zip_obj.close()
        else:
            print("No log file found.")



from math import nan

import numpy as np

DOMAIN_WISE = 0
""" Domain wise """
POINT_WISE = 1
""" Point wise """


class PredictorDataset:
    """Base (parent) class not to be used directly. Use Grib or Netcdf datasets."""

    def __init__(self, directory, file_pattern):
        self.file_pattern = file_pattern
        self.directory = directory
        self.data = []
        self.data_units = None
        self._files = None
        self.axis_lat = []
        self.axis_lon = []
        self.axis_time = []
        self.axis_level = []
        self.is_standardized = False
        self.mean = nan
        self.sd = nan

    def standardize(self, mode=DOMAIN_WISE):
        if mode == DOMAIN_WISE:
            self.mean = np.mean(self.data)
            self.sd = np.std(self.data)
            print(f"mean = {self.mean}, sd = {self.sd}")
            self.data = (self.data - self.mean) / self.sd
        elif mode == POINT_WISE:
            raise Exception("Not tested and likely not working properly.")
            self.mean = np.mean(self.data, axis=0)
            self.sd = np.std(self.data, axis=0)
            print(f"mean = {self.mean}, sd = {self.sd}")
            self.data = (self.data - self.mean) / self.sd
        else:
            raise Exception("Wrong mode for the standardization.")

        self.is_standardized = True

    def replace_nans(self, nan_val, new_val):
        self.data[self.data == nan_val] = new_val
